#! -*- coding: utf-8 -*-
"""Module to get helping methods for refractio"""

import os
import requests

from . import connection_manager


def get_conn_details_from_conn_name(connection_name, connection_type):
    """
    To get connection details by using connection_name from connection manager API.
    :param connection_name:
    :param connection_type:
    :return: connection_details
    """
    url = f"{connection_manager}/connections/api/ConnectionManager/v1/connections?" \
          f"connectionType={connection_type}&name={connection_name}"
    connection_details = requests.get(url, verify=False).json()
    return format_connection_details(connection_details)


def get_conn_details_from_ds_name(dataset_name, project_id):
    """
    To get connection details by using dataset_name and project_id from connection manager API.
    :param dataset_name:
    :param project_id:
    :return: connection_details
    """
    url = f"{connection_manager}/connections/api/External/v2/external/getConnConfig/" \
          f"{dataset_name}/refract_user/{project_id}"  # user id hard coded, as it's not being used in API code.
    return requests.get(url, verify=False).json()


def format_connection_details(con_details):
    """
    To format the connection_details in the required format.
    :param con_details:
    :return: formatted connection_details
    """
    connection_details = {}
    if con_details["connectionSources"]["sourceName"] == "RDBMS":
        connection_details = {
            "params": {
                "READER": con_details
            }
        }
    elif con_details["connectionSources"]["sourceName"] in ["CLOUD", "FILE_SYSTEMS", "FILE"]:
        connection_details = {
            "params": {
                "READER_STORAGE": con_details
            }
        }
    return connection_details


def validate_project_id(project_id):
    """
    Validates and returns project_id.
    :param project_id:
    :return: project_id
    """
    if not project_id:
        project_id = os.getenv("PROJECT_ID")
        if not project_id:
            raise Exception(f"Could not read project_id from env, please pass project_id as a keyword argument"
                            f" to get_dataframe method,\nproject_id:{project_id}")
    return project_id


def default_connection_name(datasource):
    """
    To get pick a default connection name created by the user.
    :return: connection_name
    """
    con_name = None
    url = f"{connection_manager}/connections/api/ConnectionManager/v1/connection/sourceName?profile=default"
    connection_details = requests.post(url, data=datasource, verify=False).json()

    # Reading username from OS env, will need to make a common variable in env to get current logged-in user id.
    default_user_name = os.getenv("userId") if os.getenv("userId") else (
        os.getenv("MOSAIC_ID") if os.getenv("MOSAIC_ID") else os.getenv("user_id"))

    if default_user_name:
        print(f"User name picked from OS env: {default_user_name}")
        print(f"Fetching connections created by {default_user_name} user")
        con_name = [con['name'] for con in connection_details if
                    (con["createdBy"].lower() if con["createdBy"] else None) == default_user_name.lower()]
        print(f"Connection names fetched {con_name}, created by {default_user_name}")
    else:
        print("Could not get user id from the OS env.")
    return con_name[0] if con_name else None  # first connection name from all the connections created by user_id


def get_dataframe_query(table_name, row_count, filter_condition, top=None):
    """
    To get the query to fetch dataframe
    :param table_name:
    :param row_count:
    :param filter_condition:
    :param top:
    :return: query
    """
    query = f"SELECT * FROM {table_name}"
    if top and int(row_count) > 0:
        print(f"fetching {row_count} records!")
        query = f"SELECT TOP {row_count} * FROM {table_name}"
    if filter_condition:
        query = query + " " + filter_condition
    if not top and int(row_count) > 0:
        print(f"fetching {row_count} records!")
        query = f"{query} LIMIT {row_count}"
    return query



#! -*- coding: utf-8 -*-
"""Class to get HIVE connection object and related operations"""

import os
import pandas as pd
from .manager import (
    get_conn_details_from_conn_name,
    get_conn_details_from_ds_name,
    validate_project_id,
    default_connection_name,
    get_dataframe_query
)


class Hive:

    def __init__(self):
        self.__connection_details = None
        self.con_obj = None

    @staticmethod
    def _run_os_level_command(command, switch_sudo_user=False):
        """ This function calls os level command"""
        import subprocess
        shell_type = ["sudo", "bash", "-c", command] if switch_sudo_user else ["sh", "-c", command]
        output = subprocess.run(shell_type, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
        if output.returncode != 0:
            print(output)
            return False
        return True

    def _get_connection(self, user_id):
        try:
            os.environ['USER'] = user_id
            if self.__connection_details["params"]["READER"].get("kerberosEnable"):
                print("Connecting to HIVE using Kerberos")
                if self.__connection_details["params"]["READER"].get("keytabFilePath"):
                    command = "kinit -kt {0} {1}".format(
                        self.__connection_details["params"]["READER"].get("keytabFilePath"),
                        self.__connection_details["params"]["READER"].get("userPrincipal")
                    )
                else:
                    command = "echo '{1}' | kinit {0}".format(
                        self.__connection_details["params"]["READER"].get("user") if self.__connection_details["params"]["READER"].get("user") else self.__connection_details["params"]["READER"].get("dbUserName"),
                        self.__connection_details["params"]["READER"].get("userPrincipalPassword")
                    )
                if Hive._run_os_level_command(command):
                    from refractpyhive import hive
                    os.environ["HIVE_HOSTNAME"] = self.__connection_details["params"]["READER"]["host"] if self.__connection_details["params"]["READER"].get("host") else self.__connection_details["params"]["READER"].get("ipAddress")
                    ssl_enabled = self.__connection_details["params"]["READER"]["ssl_enable"] if self.__connection_details["params"]["READER"].get("ssl_enable") else self.__connection_details["params"]["READER"].get("sslEnable")
                    self.con_obj = hive.connect(
                        # todo: To update the keys here, once we have the connection manager API created for refractio.
                        host=self.__connection_details["params"]["READER"]["host"] if self.__connection_details["params"]["READER"].get("host") else self.__connection_details["params"]["READER"].get("ipAddress"),
                        port=int(self.__connection_details["params"]["READER"]["port"]),
                        auth="KERBEROS",
                        kerberos_service_name=self.__connection_details["params"]["READER"]["sub_type"] if self.__connection_details["params"]["READER"].get("sub_type") else self.__connection_details["params"]["READER"]["connectionSources"]["connectionType"],
                        username=self.__connection_details["params"]["READER"].get("user") if self.__connection_details["params"]["READER"].get("user") else self.__connection_details["params"]["READER"].get("dbUserName"),
                        scheme="https" if ssl_enabled else "http",
                        database=self.__connection_details["params"]["READER"].get("database")
                    )
                else:
                    raise Exception(f"Exception occurred while authenticating the kerberos hive connection")
            else:
                print("Connecting to HIVE using PLAIN authentication, without Kerberos")
                from impala.dbapi import connect
                self.con_obj = connect(
                    # todo: To update the keys here, once we have the connection manager API created for refractio.
                    user=self.__connection_details["params"]["READER"].get("user") if self.__connection_details["params"]["READER"].get("user") else self.__connection_details["params"]["READER"].get("dbUserName"),
                    password=self.__connection_details["params"]["READER"].get("password") if self.__connection_details["params"]["READER"].get("password") else self.__connection_details["params"]["READER"].get("dbPassword"),
                    host=self.__connection_details["params"]["READER"]["host"] if self.__connection_details["params"]["READER"].get("host") else self.__connection_details["params"]["READER"].get("ipAddress"),
                    auth_mechanism='PLAIN',
                    port=int(self.__connection_details["params"]["READER"]["port"]),
                    database=self.__connection_details["params"]["READER"].get("database")
                )
        except Exception as ex:
            print(f"Connection details fetched: {self.__connection_details}")
            print(f"Ex: {ex}")
            raise Exception(f"Exception occurred in creating hive connection: "
                            f"{self.__connection_details.get('detailMessage')}")

    def get_connection(self, connection_name=None, user_id="1001"):
        try:
            # Getting connection.
            # connection.close() is not working in hive and there seems to be no attribute to validate whether the
            # connection is live in hive connection object.
            if self.con_obj is None:
                # Getting default connection name
                if connection_name is None:
                    connection_name = default_connection_name("HIVE")
                    if not connection_name:
                        raise Exception("Could not get the default connection name,"
                                        "please create/pass an active hive connection name.")
                self.__connection_details = get_conn_details_from_conn_name(
                    connection_name=connection_name, connection_type="hive")
                self._get_connection(user_id=user_id)
                print(f"Connection object created: {self.con_obj}\nPlease close the connection after use!")
            else:
                print(f"Existing connection object fetched: {self.con_obj}\nPlease close the connection after use!")

            return self.con_obj
        except Exception as ex:
            print(f"Exception occurred in getting hive connection: {ex}")

    def get_dataframe(self, dataset_name, project_id=os.getenv("project_id"),
                      user_id="1001", row_count=-1, filter_condition=None):
        try:
            project_id = validate_project_id(project_id)

            self.__connection_details = get_conn_details_from_ds_name(dataset_name=dataset_name, project_id=project_id)
            # Creating new connection attached to dataset_id for reading the dataset.
            self._get_connection(user_id=user_id)

            query = get_dataframe_query(self.__connection_details['params']['READER']['tables'],
                                        row_count, filter_condition)     # Get query to fetch details

            data_frame = pd.read_sql(query, self.con_obj)   # Read data from hive

            return data_frame
        except Exception as ex:
            print(f"Exception occurred in reading data_frame from hive connection: {ex}")

    def execute_query(self, query, connection_name=None, user_id="1001"):
        try:
            # Getting connection object
            if self.con_obj is None:
                self.get_connection(connection_name, user_id=user_id)

            data_frame = pd.read_sql(query, self.con_obj)   # Fetch all records in pandas dataframe

            return data_frame
        except Exception as ex:
            print(f"Exception occurred in execute_query: {ex}")

    def close_connection(self):
        self.con_obj.close()
        print("hive connection closed!")



#! -*- coding: utf-8 -*-
"""Class to get SFTP connection object and related operations"""

import os
import pandas as pd
from .manager import (
    get_conn_details_from_conn_name,
    get_conn_details_from_ds_name,
    validate_project_id,
    default_connection_name
)


class Sftp:

    def __init__(self):
        self.__connection_details = None
        self.con_obj = None

    def _get_connection(self):
        try:
            # Connect to SFTP
            import pysftp
            cnopts = pysftp.CnOpts()
            cnopts.hostkeys = None
            self.con_obj = pysftp.Connection(
                # todo: To update the keys here, once we have the connection manager API created for refractio.
                username=self.__connection_details["params"]["READER_STORAGE"].get("user") if self.__connection_details["params"]["READER_STORAGE"].get("user") else self.__connection_details["params"]["READER_STORAGE"].get("dbUserName"),
                password=self.__connection_details["params"]["READER_STORAGE"].get("password") if self.__connection_details["params"]["READER_STORAGE"].get("password") else self.__connection_details["params"]["READER_STORAGE"].get("dbPassword"),
                host=self.__connection_details["params"]["READER_STORAGE"]["host"] if self.__connection_details["params"]["READER_STORAGE"].get("host") else self.__connection_details["params"]["READER_STORAGE"].get("ipAddress"),
                cnopts=cnopts
            )
        except Exception as ex:
            print(f"Connection details fetched: {self.__connection_details}")
            print(f"Ex: {ex}")
            raise Exception(f"Exception occurred in creating sftp connection: "
                            f"{self.__connection_details.get('detailMessage')}")

    def _active_connection(self):
        try:
            if self.con_obj.listdir():
                return True
        except Exception:
            return False

    def get_connection(self, connection_name=None):
        try:
            # Getting connection
            if self.con_obj is None or not self._active_connection():
                # Getting default connection name
                if connection_name is None:
                    connection_name = default_connection_name("SFTP")
                    if not connection_name:
                        raise Exception("Could not get the default connection name,"
                                        "please create/pass an active sftp connection name.")
                self.__connection_details = get_conn_details_from_conn_name(
                    connection_name=connection_name, connection_type="sftp")
                self._get_connection()
                print(f"Connection object created: {self.con_obj}\nPlease close the connection after use!")
            else:
                print(f"Existing connection object fetched: {self.con_obj}\nPlease close the connection after use!")

            return self.con_obj
        except Exception as ex:
            print(f"Exception occurred in getting sftp connection: {ex}")

    def get_dataframe(self, dataset_name, project_id=os.getenv("project_id"), row_count=-1):
        try:
            project_id = validate_project_id(project_id)

            self.__connection_details = get_conn_details_from_ds_name(dataset_name=dataset_name, project_id=project_id)

            # Creating new connection attached to dataset_id for reading the dataset.
            self._get_connection()

            # Getting the file to current working directory
            self.con_obj.get(self.__connection_details["params"]["READER_STORAGE"]["sftp_file_path"])

            # Extracting the file name
            file_name = self.__connection_details["params"]["READER_STORAGE"]["sftp_file_path"].split("/")[-1]

            # Reading pandas dataframe
            data_frame = pd.read_csv(file_name, nrows=int(row_count)) if int(row_count) > 0 else pd.read_csv(file_name)

            self.con_obj.close()    # Closing the connection used for reading dataset.
            return data_frame
        except Exception as ex:
            print(f"Exception occurred in reading data_frame from sftp connection: {ex}")

    def close_connection(self):
        self.con_obj.close()
        print("sftp connection closed!")



import os
from .refractio import get_dataframe, get_local_dataframe
from .snowflake import Snowflake
from .mysql import Mysql
from .hive import Hive
from .sftp import Sftp
from .amazons3 import AmazonS3
from .azure import Azure
from .sqlserver import Sqlserver
from .postgres import Postgres
from .feature_store import FeastFeatureStore

# initialising constants
connection_manager = os.getenv("CONNECTION_MANAGER_BASE_URL", "http://fdc-project-manager:80/project-manager")
print(f"Connection manager service url initialised to {connection_manager}\n"
      f"If you need to update its value then update the variable CONNECTION_MANAGER_BASE_URL in os env.")

# initialising class objects
snowflake = Snowflake()
mysql = Mysql()
hive = Hive()
sftp = Sftp()
s3 = AmazonS3()
azure = Azure()
sqlserver = Sqlserver()
postgres = Postgres()

fs = FeastFeatureStore()



#! -*- coding: utf-8 -*-
"""Class to get SNOWFLAKE connection object and related operations"""

import os
from .manager import (
    get_conn_details_from_conn_name,
    get_conn_details_from_ds_name,
    validate_project_id,
    default_connection_name,
    get_dataframe_query
)


class Snowflake:

    def __init__(self):
        self.__connection_details = None
        self.con_obj = None

    def _get_connection(self):
        try:
            # Connect to Snowflake
            import snowflake.connector
            self.con_obj = snowflake.connector.connect(
                # todo: To update the keys here, once we have the connection manager API created for refractio.
                user=self.__connection_details["params"]["READER"].get("user") if self.__connection_details["params"]["READER"].get("user") else self.__connection_details["params"]["READER"].get("dbUserName"),
                password=self.__connection_details["params"]["READER"].get("password") if self.__connection_details["params"]["READER"].get("password") else self.__connection_details["params"]["READER"].get("dbPassword"),
                account=self.__connection_details["params"]["READER"].get("accountId") if self.__connection_details["params"]["READER"].get("accountId") else self.__connection_details["params"]["READER"].get("accountName"),
                database=self.__connection_details["params"]["READER"].get("database"),
                role=self.__connection_details["params"]["READER"]["role"],
                cloudPlatform=self.__connection_details["params"]["READER"]["cloudPlatform"],
                schema=self.__connection_details["params"]["READER"].get("schema"),
                wareHouse=self.__connection_details["params"]["READER"]["wareHouse"],
                region=self.__connection_details["params"]["READER"]["region"]
            )
        except Exception as ex:
            print(f"Connection details fetched: {self.__connection_details}")
            print(f"Ex: {ex}")
            raise Exception(f"Exception occurred in creating snowflake connection: "
                            f"{self.__connection_details.get('detailMessage')}")

    def get_connection(self, connection_name=None):
        try:
            # Getting connection
            if self.con_obj is None or self.con_obj.is_closed():
                # Getting default connection name
                if connection_name is None:
                    connection_name = default_connection_name("SNOWFLAKE")
                    if not connection_name:
                        raise Exception("Could not get the default connection name,"
                                        "please create/pass an active snowflake connection name.")
                self.__connection_details = get_conn_details_from_conn_name(
                    connection_name=connection_name, connection_type="snowflake")
                self._get_connection()
                print(f"Connection object created: {self.con_obj}\nPlease close the connection after use!")
            else:
                print(f"Existing connection object fetched: {self.con_obj}\nPlease close the connection after use!")

            return self.con_obj
        except Exception as ex:
            print(f"Exception occurred in getting snowflake connection: {ex}")

    def get_dataframe(self, dataset_name, project_id=os.getenv("project_id"), row_count=-1, filter_condition=None):
        try:
            project_id = validate_project_id(project_id)

            self.__connection_details = get_conn_details_from_ds_name(dataset_name=dataset_name, project_id=project_id)
            # Creating new connection attached to dataset_id for reading the dataset.
            self._get_connection()

            cur = self.con_obj.cursor()     # Creating cursor for executing query

            query = get_dataframe_query(self.__connection_details['params']['READER']['tables'],
                                        row_count, filter_condition)     # Get query to fetch details

            cur.execute(query)      # Execute query

            data_frame = cur.fetch_pandas_all()     # Fetch all records in pandas dataframe

            cur.close()
            # Closing the connection used for reading dataset.
            self.con_obj.close()
            return data_frame
        except Exception as ex:
            print(f"Exception occurred in reading data_frame from snowflake connection: {ex}")

    def execute_query(self, query, database=None, schema=None, connection_name=None):
        try:
            # Getting connection object
            if self.con_obj is None or self.con_obj.is_closed():
                self.get_connection(connection_name)

            # Creating cursor for executing query
            cur = self.con_obj.cursor()
            if database and schema:
                cur.execute(f"USE {database}.{schema}")     # To select database and schema

            cur.execute(query)  # Execute user query

            data_frame = cur.fetch_pandas_all()     # Fetch all records in pandas dataframe

            cur.close()     # Closing the cursor
            self.con_obj.close()    # Closing the connection
            return data_frame
        except Exception as ex:
            print(f"Exception occurred in execute_query: {ex}")

    def close_connection(self):
        self.con_obj.close()
        print("Snowflake connection closed!")



#! -*- coding: utf-8 -*-
"""Class to get Azure connection object and related operations"""

import os
from io import StringIO
import pandas as pd
from .manager import (
    get_conn_details_from_conn_name,
    get_conn_details_from_ds_name,
    validate_project_id,
    default_connection_name
)


class Azure:

    def __init__(self):
        self.__connection_details = None
        self.con_obj = None

    def _get_connection(self):
        try:
            # Connect to Azure Blob
            from azure.storage.blob import BlockBlobService
            # create a BlockBlobService object
            self.con_obj = BlockBlobService(
                account_name=self.__connection_details["params"]["READER_STORAGE"]["account_name"] if self.__connection_details["params"]["READER_STORAGE"].get("account_name") else self.__connection_details["params"]["READER_STORAGE"]["accountName"],
                account_key=self.__connection_details["params"]["READER_STORAGE"]["account_key"] if self.__connection_details["params"]["READER_STORAGE"].get("account_key") else self.__connection_details["params"]["READER_STORAGE"]["accountKey"]
            )
        except Exception as ex:
            print(f"Connection details fetched: {self.__connection_details}")
            print(f"Ex: {ex}")
            raise Exception(f"Exception occurred in creating azure blob connection: "
                            f"{self.__connection_details.get('detailMessage')}")

    def get_connection(self, connection_name=None):
        try:
            # Getting connection
            if self.con_obj is None:
                # Getting default connection name
                if connection_name is None:
                    connection_name = default_connection_name("AZURE")
                    if not connection_name:
                        raise Exception("Could not get the default connection name,"
                                        "please create/pass an active azure blob connection name.")
                self.__connection_details = get_conn_details_from_conn_name(
                    connection_name=connection_name, connection_type="azure")
                self._get_connection()
                print(f"Connection object created: {self.con_obj}")
            else:
                print(f"Existing connection object fetched: {self.con_obj}")

            return self.con_obj
        except Exception as ex:
            print(f"Exception occurred in getting azure blob connection: {ex}")

    def get_dataframe(self, dataset_name, project_id=os.getenv("project_id"), row_count=-1):
        try:
            project_id = validate_project_id(project_id)

            self.__connection_details = get_conn_details_from_ds_name(dataset_name=dataset_name, project_id=project_id)

            # Creating new connection attached to dataset_id for reading the dataset.
            self._get_connection()

            # create a container
            container_name = self.__connection_details["params"]["READER_STORAGE"]["container"]
            self.con_obj.create_container(container_name)

            # get the data from the blob
            blob_name = self.__connection_details["params"]["READER_STORAGE"]["blob_name"]
            blob_data = self.con_obj.get_blob_to_bytes(container_name, blob_name)

            s = str(blob_data.content, 'utf-8')
            data = StringIO(s)
            return pd.read_csv(data, nrows=int(row_count)) if int(row_count) > 0 else pd.read_csv(data)
        except Exception as ex:
            print(f"Exception occurred in reading data_frame from azure blob connection: {ex}")



#! -*- coding: utf-8 -*-
"""Class to get AmazonS3 connection object and related operations"""

import os
import pandas as pd
from .manager import (
    get_conn_details_from_conn_name,
    get_conn_details_from_ds_name,
    validate_project_id,
    default_connection_name
)


class AmazonS3:

    def __init__(self):
        self.__connection_details = None
        self.con_obj = None

    def _get_connection(self):
        try:
            # Connect to AmazonS3
            import boto3
            self.con_obj = boto3.client(
                "s3",
                aws_access_key_id=self.__connection_details["params"]["READER_STORAGE"]["access_key"] if self.__connection_details["params"]["READER_STORAGE"].get("access_key") else self.__connection_details["params"]["READER_STORAGE"]["accessKey"],
                aws_secret_access_key=self.__connection_details["params"]["READER_STORAGE"]["secret_key"] if self.__connection_details["params"]["READER_STORAGE"].get("secret_key") else self.__connection_details["params"]["READER_STORAGE"]["secretKey"]
            )
        except Exception as ex:
            print(f"Connection details fetched: {self.__connection_details}")
            print(f"Ex: {ex}")
            raise Exception(f"Exception occurred in creating amazon s3 connection: "
                            f"{self.__connection_details.get('detailMessage')}")

    def get_connection(self, connection_name=None):
        try:
            # Getting connection
            if self.con_obj is None:
                # Getting default connection name
                if connection_name is None:
                    connection_name = default_connection_name("AMAZONS3")
                    if not connection_name:
                        raise Exception("Could not get the default connection name,"
                                        "please create/pass an active amazon s3 connection name.")
                self.__connection_details = get_conn_details_from_conn_name(
                    connection_name=connection_name, connection_type="amazons3")
                self._get_connection()
                print(f"Connection object created: {self.con_obj}")
            else:
                print(f"Existing connection object fetched: {self.con_obj}")

            return self.con_obj
        except Exception as ex:
            print(f"Exception occurred in getting amazon s3 connection: {ex}")

    def get_dataframe(self, dataset_name, project_id=os.getenv("project_id"), row_count=-1):
        try:
            project_id = validate_project_id(project_id)

            self.__connection_details = get_conn_details_from_ds_name(dataset_name=dataset_name, project_id=project_id)

            # Creating new connection attached to dataset_id for reading the dataset.
            self._get_connection()

            file_name = self.__connection_details["params"]["READER_STORAGE"]["s3_folder"] + "/" + \
                        self.__connection_details["params"]["READER_STORAGE"]["file"]
            response = self.con_obj.get_object(Bucket=self.__connection_details["params"]["READER_STORAGE"][
                "bucket_name"], Key=file_name)
            status = response.get("ResponseMetadata", {}).get("HTTPStatusCode")

            if status == 200:
                print(f"Successful S3 get_object response. Status - {status}")
                return pd.read_csv(response.get("Body"), nrows=int(row_count)) if int(
                    row_count) > 0 else pd.read_csv(response.get("Body"))
            else:
                print(f"Unsuccessful S3 get_object response. Status - {status}")
        except Exception as ex:
            print(f"Exception occurred in reading data_frame from amazon s3 connection: {ex}")



#! -*- coding: utf-8 -*-
"""Class to get postgres connection object and related operations"""

import os
import pandas.io.sql as sqlio
from .manager import (
    get_conn_details_from_conn_name,
    get_conn_details_from_ds_name,
    validate_project_id,
    default_connection_name,
    get_dataframe_query
)


class Postgres:

    def __init__(self):
        self.__connection_details = None
        self.con_obj = None

    def _get_connection(self):
        try:
            # Connect to Postgres
            import psycopg2
            self.con_obj = psycopg2.connect(
                # todo: To update the keys here, once we have the connection manager API created for refractio.
                user=self.__connection_details["params"]["READER"].get("user") if self.__connection_details["params"]["READER"].get("user") else self.__connection_details["params"]["READER"].get("dbUserName"),
                password=self.__connection_details["params"]["READER"].get("password") if self.__connection_details["params"]["READER"].get("password") else self.__connection_details["params"]["READER"].get("dbPassword"),
                host=self.__connection_details["params"]["READER"]["host"] if self.__connection_details["params"]["READER"].get("host") else self.__connection_details["params"]["READER"].get("ipAddress"),
                port=int(self.__connection_details["params"]["READER"]["port"]),
                database=self.__connection_details["params"]["READER"].get("database") if self.__connection_details["params"]["READER"].get("database") else self.__connection_details["params"]["READER"].get("defaultDb")
            )
        except Exception as ex:
            print(f"Connection details fetched: {self.__connection_details}")
            print(f"Ex: {ex}")
            raise Exception(f"Exception occurred in creating postgres connection: "
                            f"{self.__connection_details.get('detailMessage')}")

    def get_connection(self, connection_name=None):
        try:
            # Getting connection
            if self.con_obj is None or self.con_obj.closed:
                # Getting default connection name
                if connection_name is None:
                    connection_name = default_connection_name("POSTGRES")
                    if not connection_name:
                        raise Exception("Could not get the default connection name,"
                                        "please create/pass an active postgres connection name.")
                self.__connection_details = get_conn_details_from_conn_name(
                    connection_name=connection_name, connection_type="postgres")
                self._get_connection()
                print(f"Connection object created: {self.con_obj}\nPlease close the connection after use!")
            else:
                print(f"Existing connection object fetched: {self.con_obj}\nPlease close the connection after use!")

            return self.con_obj
        except Exception as ex:
            print(f"Exception occurred in getting postgres connection: {ex}")

    def get_dataframe(self, dataset_name, project_id=os.getenv("project_id"), row_count=-1, filter_condition=None):
        try:
            project_id = validate_project_id(project_id)

            self.__connection_details = get_conn_details_from_ds_name(dataset_name=dataset_name, project_id=project_id)
            # Creating new connection attached to dataset_id for reading the dataset.
            self._get_connection()

            query = get_dataframe_query(f"{self.__connection_details['params']['READER']['schema']}."
                                        f"{self.__connection_details['params']['READER']['tables']}",
                                        row_count, filter_condition)     # Get query to fetch details

            data_frame = sqlio.read_sql_query(query, con=self.con_obj)   # Read data from postgres

            self.con_obj.close()    # Closing the connection used for reading dataset.
            return data_frame
        except Exception as ex:
            print(f"Exception occurred in reading data_frame from postgres connection: {ex}")

    def execute_query(self, query, connection_name=None):
        try:
            # Getting connection object
            if self.con_obj is None or self.con_obj.closed:
                self.get_connection(connection_name)

            data_frame = sqlio.read_sql_query(query, con=self.con_obj)   # Fetch all records in pandas dataframe

            self.con_obj.close()    # Closing the connection
            return data_frame
        except Exception as ex:
            print(f"Exception occurred in execute_query: {ex}")

    def close_connection(self):
        self.con_obj.close()
        print("postgres connection closed!")



from setuptools import setup, find_packages

# read the contents of README file
from pathlib import Path
this_directory = Path(__file__).parent
LONG_DESCRIPTION = (this_directory / "README.md").read_text()

VERSION = '2.1.5.1'
DESCRIPTION = 'REFRACT-IO: To read and write dataframe from different connectors.'

extras_require = {
    "all": [
        "snowflake-connector-python[pandas]==3.6.0",
        "boto3==1.26.116",
        "azure==4.0.0",
        "openpyxl==3.1.2",
        "xlrd==2.0.1",
        "pysftp==0.2.9",
        "pymysql==1.0.3",
        "impyla==0.16.2",
        "thrift-sasl==0.4.3",
        "pure-sasl==0.6.2",
        "PyHive==0.6.3.dev0",
        "pyodbc==4.0.39",
        "psycopg2-binary==2.9.6"
    ],
    "snowflake": [
        "snowflake-connector-python[pandas]==3.6.0"
    ],
    "s3": [
        "boto3==1.26.116"
    ],
    "azureblob": [
        "azure==4.0.0"
    ],
    "local": [
        "openpyxl==3.1.2",
        "xlrd==2.0.1",
    ],
    "sftp": [
        "pysftp==0.2.9"
    ],
    "mysql": [
        "pymysql==1.0.3"
    ],
    "hive": [
        "impyla==0.16.2",
        "thrift-sasl==0.4.3",
        "pure-sasl==0.6.2",
        "PyHive==0.6.3.dev0",
        "refractpyhive==0.0.3.dev0"
    ],
    "sqlserver": [
        "pyodbc==4.0.39"
    ],
    "postgres": [
        "psycopg2-binary==2.9.6"
    ]
}

# Setting up
setup(
    name="refractio",
    version=VERSION,
    author="Abhishek Chaurasia",
    author_email="<abhishek1.chaurasia@fosfor.com>",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "pandas==2.0.0"
    ],
    keywords=['refractio'],
    classifiers=[
        "Programming Language :: Python :: 3.8",
    ],
    extras_require=extras_require,
    project_urls={
        "Product": "https://www.fosfor.com/refract/",
        "Source": "https://gitlab.fosfor.com/fosfor-decision-cloud/intelligence/refract-sdk",
    }
)



#! -*- coding: utf-8 -*-
"""Library to read and write dataframes"""

import os
import requests
import pandas as pd

from .manager import validate_project_id, get_dataframe_query
from . import connection_manager


def get_dataframe(ds_name, project_id=os.getenv("project_id"), row_count=-1, strategy="top", user_id="1001",
                  filter_condition=None):
    """
    Param:
        ds_name,
        project_id=os.getenv("project_id"),
        row_count=-1,
        strategy="top"
        user_id="1001"
    To get pandas dataframe.
    Need to install mosaic-connector-python for reading dataframe using connector backend,
    git+https://gitlab+deploy-token-14:myUpFE_XRxShG53Hs6tV@git.lti-aiq.in/mosaic-decisions-2-0/mosaic-connector-python.git@1.0.29.3
    """
    try:
        project_id = validate_project_id(project_id)
        url = f"{connection_manager}/connections/api/External/v2/external/getConnConfig/" \
              f"{ds_name}/refract_user/{project_id}"    # user id hard coded, as it's not being used in API code.
        connection_details = requests.get(url, verify=False).json()

        if connection_details["params"]["READER"]["type"] == "RDBMS":
            if connection_details["params"]["READER"]["sub_type"] == "SNOWFLAKE":
                data_frame = get_snowflake_df(connection_details, row_count, strategy, filter_condition)
            elif connection_details["params"]["READER"]["sub_type"] == "MYSQL":
                data_frame = get_mysql_df(connection_details, row_count, strategy, filter_condition)
            elif connection_details["params"]["READER"]["sub_type"] == "HIVE":
                data_frame = get_hive_df(connection_details, project_id, row_count, strategy, user_id, filter_condition)
            elif connection_details["params"]["READER"]["sub_type"] == "SQLSERVER":
                data_frame = get_sqlserver_df(connection_details, row_count, strategy, filter_condition)
            elif connection_details["params"]["READER"]["sub_type"] == "POSTGRES":
                data_frame = get_postgres_df(connection_details, row_count, strategy, filter_condition)
            else:
                print("Reading dataframe using connector backend")
                # Need to install, git+https://gitlab+deploy-token-14:myUpFE_XRxShG53Hs6tV@git.lti-aiq.in/mosaic-decisions-2-0/mosaic-connector-python.git
                from connector.mosaicio import MosaicioConnector
                connector = MosaicioConnector()
                data_frame = connector.getPandasDataFrame(
                    param=connection_details["params"],
                    row_count=row_count,
                    strategy=strategy
                )
        elif connection_details["params"]["READER"]["type"] == "FILE":
            if connection_details["params"]["READER_STORAGE"]["type"] == "AMAZONS3":
                data_frame = get_s3_df(connection_details, row_count, strategy)
            elif connection_details["params"]["READER_STORAGE"]["type"] == "AZURE":
                data_frame = get_azureblob_df(connection_details, row_count, strategy)
            elif connection_details["params"]["READER_STORAGE"]["type"] == "SFTP":
                data_frame = get_sftp_df(connection_details, row_count, strategy)
        return data_frame
    except Exception as ex:
        print(f'project_id: {project_id}')
        print(f"Connection details fetched: {connection_details}")
        print(f"Exception occurred in get_dataframe: {ex}")


def get_local_dataframe(local_file_path, row_count=-1):
    """
    Param:
        local_file_path,
        row_count
    To read data frame from NAS local file system
    """
    try:
        sub_type = local_file_path.split(".")[-1]
        nrows = row_count if int(row_count) > 0 else None
        if sub_type == "csv":
            return pd.read_csv(filepath_or_buffer=local_file_path,
                               sep=",",
                               nrows=nrows)
        elif sub_type == "tsv":
            return pd.read_csv(filepath_or_buffer=local_file_path,
                               sep="\t",
                               nrows=nrows)
        elif sub_type == "xlsx":
            return pd.read_excel(io=local_file_path,
                                 nrows=nrows,
                                 parse_dates=False,
                                 engine="openpyxl")
        elif sub_type == "xls":
            return pd.read_excel(io=local_file_path,
                                 nrows=nrows,
                                 parse_dates=False)
    except Exception as ex:
        print(f"Exception occurred in get_local_dataframe: {ex}")


def get_snowflake_df(connection_details, row_count, strategy, filter_condition):
    """
    Param:
        connection_details: connection details dict
        row_count: number of rows to be fetched
        strategy: top/bottom
    To read data frame form snowflake connection
    """
    print("Reading dataframe from snowflake native connector")

    import snowflake.connector

    # Connect to Snowflake
    con = snowflake.connector.connect(
        user=connection_details["params"]["READER"]["user"],
        password=connection_details["params"]["READER"]["password"],
        account=connection_details["params"]["READER"]["accountId"],
        database=connection_details["params"]["READER"]["database"],
        role=connection_details["params"]["READER"]["role"],
        cloudPlatform=connection_details["params"]["READER"]["cloudPlatform"],
        schema=connection_details["params"]["READER"]["schema"],
        wareHouse=connection_details["params"]["READER"]["wareHouse"],
        region=connection_details["params"]["READER"]["region"]
    )
    # Create cursor
    cur = con.cursor()
    query = get_dataframe_query(connection_details['params']['READER']['tables'],
                                row_count, filter_condition)  # Get query to fetch details

    cur.execute(query)  # Execute query

    # Read results into a pandas DataFrame
    data_frame = cur.fetch_pandas_all()

    # Close cursor and connection
    cur.close()
    con.close()
    return data_frame


def get_hive_df(connection_details, project_id, row_count, strategy, user_id, filter_condition):
    """
    Param:
        connection_details: connection details dict
        row_count: number of rows to be fetched
        strategy: top/bottom
        user_id: template image user id
    To read data frame form hive connection
    """
    print("Reading dataframe from hive native connector")

    from . import hive
    data_frame = hive.get_dataframe(
        dataset_name=connection_details['params']['READER']['tables'],
        project_id=project_id,
        user_id=user_id,
        row_count=row_count,
        filter_condition=filter_condition
    )
    hive.close_connection()

    return data_frame


def get_mysql_df(connection_details, row_count, strategy, filter_condition):
    """
    Param:
        connection_details: connection details dict
        row_count: number of rows to be fetched
        strategy: top/bottom
    To read data frame form mysql connection
    """
    print("Reading dataframe from mysql native connector")
    import pymysql

    conn = pymysql.connect(host=connection_details["params"]["READER"]["host"],
                           port=int(connection_details["params"]["READER"]["port"]),
                           user=connection_details["params"]["READER"]["user"],
                           passwd=connection_details["params"]["READER"]["password"],
                           db=connection_details["params"]["READER"]["database"])

    query = get_dataframe_query(connection_details['params']['READER']['tables'],
                                row_count, filter_condition)  # Get query to fetch details
    # Read data fromm mysql
    data_frame = pd.read_sql(query, con=conn)
    # Close connection
    conn.close()
    return data_frame


def get_sqlserver_df(connection_details, row_count, strategy, filter_condition):
    """
    Param:
        connection_details: connection details dict
        row_count: number of rows to be fetched
        strategy: top/bottom
    To read data frame from sql server connection
    """
    print("Reading dataframe from sqlserver native connector")
    import pyodbc

    # This need the OS should have sql server driver installed. Like,
    # curl https://packages.microsoft.com/config/rhel/9.0/prod.repo > /etc/yum.repos.d/mssql-release.repo
    # ACCEPT_EULA=Y yum install -y msodbcsql18
    # driver_lib = "/opt/microsoft/msodbcsql18/lib64/libmsodbcsql-18.2.so.1.1"    # sqlserver driver lib path
    driver_lib = "{ODBC Driver 18 for SQL Server}"  # sqlserver driver version instead of lib path
    conn = pyodbc.connect(f'DRIVER={driver_lib};'
                          f'SERVER={connection_details["params"]["READER"]["host"]};'
                          f'DATABASE={connection_details["params"]["READER"]["database"]};'
                          f'UID={connection_details["params"]["READER"]["user"]};'
                          f'PWD={connection_details["params"]["READER"]["password"]};'
                          f'TrustServerCertificate=yes;')

    query = get_dataframe_query(connection_details['params']['READER']['tables'],
                                row_count, filter_condition, top=True)  # Get query to fetch details
    # Read data from sqlserver
    data_frame = pd.read_sql(query, con=conn)
    # Close connection
    conn.close()
    return data_frame


def get_postgres_df(connection_details, row_count, strategy, filter_condition):
    """
    Param:
        connection_details: connection details dict
        row_count: number of rows to be fetched
        strategy: top/bottom
    To read data frame form postgres connection
    """
    print("Reading dataframe from postgres native connector")
    import psycopg2
    import pandas.io.sql as sqlio

    conn = psycopg2.connect(host=connection_details["params"]["READER"]["host"],
                            port=int(connection_details["params"]["READER"]["port"]),
                            user=connection_details["params"]["READER"]["user"],
                            password=connection_details["params"]["READER"]["password"],
                            database=connection_details["params"]["READER"]["database"])

    query = get_dataframe_query(f"{connection_details['params']['READER']['schema']}."
                                f"{connection_details['params']['READER']['tables']}",
                                row_count, filter_condition)  # Get query to fetch details
    # Read data fromm postgres
    data_frame = sqlio.read_sql_query(query, con=conn)
    # Close connection
    conn.close()
    return data_frame


def get_s3_df(connection_details, row_count, strategy):
    """
    Param:
        connection_details: connection details dict
        row_count: number of rows to be fetched
        strategy: top/bottom
    To read data frame form amazon S3 connection
    """
    print("Reading dataframe from s3 native connector")

    import boto3

    s3_client = boto3.client(
        "s3",
        aws_access_key_id=connection_details["params"]["READER_STORAGE"]["access_key"],
        aws_secret_access_key=connection_details["params"]["READER_STORAGE"]["secret_key"],
    )
    file_name = connection_details["params"]["READER_STORAGE"]["s3_folder"] + "/" + \
                connection_details["params"]["READER_STORAGE"]["file"]
    response = s3_client.get_object(Bucket=connection_details["params"]["READER_STORAGE"]["bucket_name"],
                                    Key=file_name)
    status = response.get("ResponseMetadata", {}).get("HTTPStatusCode")

    if status == 200:
        print(f"Successful S3 get_object response. Status - {status}")
        data_frame = pd.read_csv(response.get("Body"), nrows=int(row_count)) if int(row_count) > 0 else pd.read_csv(
            response.get("Body"))
        return data_frame
    else:
        print(f"Unsuccessful S3 get_object response. Status - {status}")


def get_azureblob_df(connection_details, row_count, strategy):
    """
    Param:
        connection_details: connection details dict
        row_count: number of rows to be fetched
        strategy: top/bottom
    To read data frame form azureblob connection
    """
    print("Reading dataframe from azureblob native connector")

    from azure.storage.blob import BlockBlobService
    from io import StringIO

    # create a BlockBlobService object
    blob_service = BlockBlobService(
        account_name=connection_details["params"]["READER_STORAGE"]["account_name"],
        account_key=connection_details["params"]["READER_STORAGE"]["account_key"]
    )

    # create a container
    container_name = connection_details["params"]["READER_STORAGE"]["container"]
    blob_service.create_container(container_name)

    # get the data from the blob
    blob_name = connection_details["params"]["READER_STORAGE"]["blob_name"]
    blob_data = blob_service.get_blob_to_bytes(container_name, blob_name)

    s = str(blob_data.content, 'utf-8')
    data = StringIO(s)
    # read the data
    data_frame = pd.read_csv(data, nrows=int(row_count)) if int(row_count) > 0 else pd.read_csv(data)
    return data_frame


def get_sftp_df(connection_details, row_count, strategy):
    """
    Param:
        connection_details: connection details dict
        row_count: number of rows to be fetched
        strategy: top/bottom
    To read data frame form sftp connection
    """
    print("Reading dataframe from sftp native connector")

    import pysftp
    cnopts = pysftp.CnOpts()
    cnopts.hostkeys = None
    with pysftp.Connection(host=connection_details["params"]["READER_STORAGE"]["host"],
                           username=connection_details["params"]["READER_STORAGE"]["user"],
                           password=connection_details["params"]["READER_STORAGE"]["password"],
                           cnopts=cnopts) as sftp:
        sftp.get(connection_details["params"]["READER_STORAGE"]["sftp_file_path"])
        file_name = connection_details["params"]["READER_STORAGE"]["sftp_file_path"].split("/")[-1]
        data_frame = pd.read_csv(file_name, nrows=int(row_count)) if int(row_count) > 0 else pd.read_csv(file_name)
        return data_frame


def write_rdbms_dataframe(data_frame, connection_id, data_source, database_name, schema, table_name):
    """
    Param:
        data_frame: data_frame to write to RDBMS
        connection_id: user selected connection_id
        data_source: HIVE, MYSQL, ORACLE, SNOWFLAKE, SQLSERVER, POSTGRES
        database_name: user selected DB name
        schema: user selected schema
        table_name: output table name
    To write data frame using connectors backend code.
    """
    try:
        url = f"{connection_manager}/connections/api/ConnectionManager/v1/connection/{connection_id}"
        connection_details = requests.get(url, data=data_source, verify=False).json()
        writer = {
            "WRITER": {
                "sub_type": connection_details["connectionSources"]["connectionType"],
                "type": connection_details["connectionSources"]["sourceName"],
                "user": connection_details["dbUserName"],
                "password": connection_details["dbPassword"],
                "port": connection_details["port"],
                "host": connection_details["ipAddress"],
                "database": database_name,
                "schema": schema,
                "tables": table_name,
                "DATASET_WRITE_MODE": "overwrite",
                "accountId": connection_details.get("accountName"),
                "region": connection_details.get("region"),
                "cloudPlatform": connection_details.get("cloudPlatform"),
                "wareHouse": connection_details.get("wareHouse")
            }
        }
        from connector.mosaicio import MosaicioConnector
        MosaicioConnector().writePandasDataFrame(data_frame, writer)
    except Exception as ex:
        print(f"Exception occurred in write_rdbms_dataframe: {ex}")



import os
import requests
import tempfile

class FeastFeatureStore:
    def get_feature_store(self, feature_store_name, project_id):
        try:
            from feast import FeatureStore
            headers = {
                "accept": "application/json",
                "X-Project-Id": project_id,
                'X-Auth-Userid': os.getenv("userId"),
                'X-Auth-Username': os.getenv("userId"),
                'X-Auth-Email': os.getenv("userId"),
            }
            base_url = "http://refract-common-service:5000/refract/common/api"
            url = f"{base_url}/v1/get_feature_store?feature_store_name={feature_store_name}"
            response = requests.get(url=url, headers=headers, verify=False)
            if response.status_code == 403:
                final_dict = response.json()
                error_message = final_dict.get('error', 'Authorization error')
                raise AuthorizationException(error_message)

            temp_dir = tempfile.mkdtemp()
            yaml_path = os.path.join(temp_dir, "feature_store.yaml")
            if response.status_code == 200:
                with open(yaml_path, 'wb') as f:
                    f.write(response.content)
                store = FeatureStore(repo_path=temp_dir)
                return store
        except Exception as e:
            print(f"Error: {e}")


class AuthorizationException(Exception):
    def __init__(self, message="Authorization error"):
        self.message = message
        super().__init__(self.message)
